/*
Conference.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import androidx.annotation.Nullable;
import androidx.annotation.NonNull;

/**
* TODO. <br/>
* <br/>
*/
public interface Conference {
    public enum State {
        /**
        * Initial state. <br/>
        * <br/>
        */
        None(0),

        /**
        * Conference is now instantiated on local. <br/>
        * <br/>
        */
        Instantiated(1),

        /**
        * One creation request was sent to the server. <br/>
        * <br/>
        */
        CreationPending(2),

        /**
        * Conference was created on the server. <br/>
        * <br/>
        */
        Created(3),

        /**
        * Conference creation failed. <br/>
        * <br/>
        */
        CreationFailed(4),

        /**
        * Wait for conference termination. <br/>
        * <br/>
        */
        TerminationPending(5),

        /**
        * Conference exists on server but not in local. <br/>
        * <br/>
        */
        Terminated(6),

        /**
        * Conference termination failed. <br/>
        * <br/>
        */
        TerminationFailed(7),

        /**
        * Conference was deleted on the server. <br/>
        * <br/>
        */
        Deleted(8);

        protected final int mValue;

        private State (int value) {
            mValue = value;
        }

        static public State fromInt(int value) throws RuntimeException {
            switch(value) {
            case 0: return None;
            case 1: return Instantiated;
            case 2: return CreationPending;
            case 3: return Created;
            case 4: return CreationFailed;
            case 5: return TerminationPending;
            case 6: return Terminated;
            case 7: return TerminationFailed;
            case 8: return Deleted;
            default:
                throw new RuntimeException("Unhandled enum value " + value + " for State");
            }
        }

        public int toInt() {
            return mValue;
        }
    };

    /**
    * Get the conference address of the conference. <br/>
    * <br/>
    * @return The conference address of the conference or null if it is unknown.   <br/>
    */
    @Nullable
    public Address getConferenceAddress();

    /**
    * Returns core for a {@link Conference}. <br/>
    * <br/>
    * @return back pointer to {@link Core} object.   Returns back pointer to {@link Core}<br/>
    * object. <br/>
    */
    @NonNull
    public Core getCore();

    /**
    * Get current parameters of the conference. <br/>
    * <br/>
    * @return a {@link ConferenceParams} .   <br/>
    */
    @NonNull
    public ConferenceParams getCurrentParams();

    /**
    * Get the conference id as string. <br/>
    * <br/>
    * @return the conference id   <br/>
    */
    @Nullable
    public String getId();

    /**
    * Set the conference id as string. <br/>
    * <br/>
    * @param conferenceId the conference id to set.   <br/>
    */
    public void setId(@Nullable String conferenceId);

    /**
    * For a local audio video conference, this function returns the participant<br/>
    * hosting the conference For a remote audio video conference, this function<br/>
    * returns the focus of the conference. <br/>
    * <br/>
    * @return a {@link Participant} .   <br/>
    */
    @NonNull
    public Participant getMe();

    /**
    * Get number of participants. <br/>
    * <br/>
    * @return the number of participants in a {@link Conference} <br/>
    */
    public int getParticipantCount();

    /**
    * Get list of all participants of one conference. <br/>
    * <br/>
    * @return A list of participants.      <br/>
    */
    @NonNull
    public Participant[] getParticipantList();

    /**
    * Get URIs of all participants of one conference The returned bctbx_list_t<br/>
    * contains URIs of all participant. <br/>
    * <br/>
    * That list must be freed after use and each URI must be unref with<br/>
    * linphone_address_unref <br/>
    * @return The list of the participants' address.      <br/>
    * @deprecated 10/07/2020 Use {@link #getParticipantList} instead. <br/>
    */
    @Deprecated
    @NonNull
    public Address[] getParticipants();

    /**
    * Get the conference subject. <br/>
    * <br/>
    * @return conference subject.   <br/>
    */
    @Nullable
    public String getSubject();

    /**
    * Set the conference subject. <br/>
    * <br/>
    * @param subject conference subject   <br/>
    */
    public void setSubject(@Nullable String subject);

    /**
    * Join a participant to the conference. <br/>
    * <br/>
    * @param uri a {@link Address} that has to be added to the conference.   <br/>
    */
    public int addParticipant(@NonNull Address uri);

    /**
    * Join an existing call to the conference. <br/>
    * <br/>
    * @param call a {@link Call} that has to be added to the conference.   <br/>
    */
    public int addParticipant(@NonNull Call call);

    /**
    * Find a participant from a conference. <br/>
    * <br/>
    * @param uri SIP URI of the participant to search.   <br/>
    * @return a pointer to the participant found or nullptr.   <br/>
    */
    @Nullable
    public Participant findParticipant(@NonNull Address uri);

    /**
    * Invite participants to the conference, by supplying a list of {@link Address}. <br/>
    * <br/>
    * @param addresses A list of SIP addresses to invite.    <br/>
    * @param params {@link CallParams} to use for inviting the participants.   <br/>
    */
    public int inviteParticipants(@NonNull Address[] addresses, @Nullable CallParams params);

    /**
    * Call generic OpenGL render for a given conference. <br/>
    * <br/>
    */
    public void oglRender();

    /**
    * Call generic OpenGL render preview for a given conference. <br/>
    * <br/>
    */
    public void previewOglRender();

    /**
    * <br/>
    * @param call call to remove   <br/>
    * @return 0 if succeeded, -1 if failed <br/>
    * @deprecated 10/07/2020 Use {@link #removeParticipant} instead. <br/>
    */
    @Deprecated
    public int removeParticipant(@NonNull Call call);

    /**
    * <br/>
    * @param participant participant to remove   <br/>
    * warning: The passed participant must be one of those returned by {@link #getParticipantList}<br/>
    * warning: This method may destroy the conference if the only remaining<br/>
    * participant had an existing call to the local participant before the conference<br/>
    * was created <br/>
    * @return 0 if succeeded, -1 if failed <br/>
    */
    public int removeParticipant(@NonNull Participant participant);

    /**
    * <br/>
    * @param uri URI of the participant to remove   <br/>
    * warning: The passed participant uri must be one of those returned by {@link #getParticipants}<br/>
    * @return 0 if succeeded, -1 if failed <br/>
    * @deprecated 10/07/2020 Use {@link #removeParticipant} instead. <br/>
    */
    @Deprecated
    public int removeParticipant(@NonNull Address uri);

    /**
    * Terminates conference. <br/>
    * <br/>
    * @return 0 if the termination is successful, -1 otherwise. <br/>
    */
    public int terminate();

    /**
    * Update parameters of the conference. <br/>
    * <br/>
    * This is typically used enable or disable the video stream in the conference. <br/>
    * @param params the new parameters to apply.   <br/>
    */
    public int updateParams(@NonNull ConferenceParams params);

    /**
    */
    public void addListener(ConferenceListener listener);

    /**
    */
    public void removeListener(ConferenceListener listener);

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();

    /**
      * Gets the native pointer used by this class to make native method calls.
      * @return the nativer pointer, as long
      */
    public long getNativePointer();

    public String toString();
}

class ConferenceImpl implements Conference {

    protected long nativePtr = 0;
    protected transient Object userData = null;

    protected ConferenceImpl(long ptr) {
        nativePtr = ptr;
    }

    public long getNativePointer() {
        return nativePtr;
    }


    private native Address getConferenceAddress(long nativePtr);
    @Override @Nullable
    synchronized public Address getConferenceAddress()  {
        
        return (Address)getConferenceAddress(nativePtr);
    }

    private native Core getCore(long nativePtr);
    @Override @NonNull
    synchronized public Core getCore()  {
        
        return (Core)getCore(nativePtr);
    }

    private native ConferenceParams getCurrentParams(long nativePtr);
    @Override @NonNull
    synchronized public ConferenceParams getCurrentParams()  {
        
        return (ConferenceParams)getCurrentParams(nativePtr);
    }

    private native String getId(long nativePtr);
    @Override @Nullable
    synchronized public String getId()  {
        
        return getId(nativePtr);
    }

    private native void setId(long nativePtr, String conferenceId);
    @Override
    synchronized public void setId(@Nullable String conferenceId)  {
        
        setId(nativePtr, conferenceId);
    }

    private native Participant getMe(long nativePtr);
    @Override @NonNull
    synchronized public Participant getMe()  {
        
        return (Participant)getMe(nativePtr);
    }

    private native int getParticipantCount(long nativePtr);
    @Override
    synchronized public int getParticipantCount()  {
        
        return getParticipantCount(nativePtr);
    }

    private native Participant[] getParticipantList(long nativePtr);
    @Override @NonNull
    synchronized public Participant[] getParticipantList()  {
        
        return getParticipantList(nativePtr);
    }

    private native Address[] getParticipants(long nativePtr);
    @Override @NonNull
    synchronized public Address[] getParticipants()  {
        
        return getParticipants(nativePtr);
    }

    private native String getSubject(long nativePtr);
    @Override @Nullable
    synchronized public String getSubject()  {
        
        return getSubject(nativePtr);
    }

    private native void setSubject(long nativePtr, String subject);
    @Override
    synchronized public void setSubject(@Nullable String subject)  {
        
        setSubject(nativePtr, subject);
    }

    private native int addParticipant2(long nativePtr, Address uri);
    @Override
    synchronized public int addParticipant(@NonNull Address uri)  {
        
        return addParticipant2(nativePtr, uri);
    }

    private native int addParticipant(long nativePtr, Call call);
    @Override
    synchronized public int addParticipant(@NonNull Call call)  {
        
        return addParticipant(nativePtr, call);
    }

    private native Participant findParticipant(long nativePtr, Address uri);
    @Override @Nullable
    synchronized public Participant findParticipant(@NonNull Address uri)  {
        
        return (Participant)findParticipant(nativePtr, uri);
    }

    private native int inviteParticipants(long nativePtr, Address[] addresses, CallParams params);
    @Override
    synchronized public int inviteParticipants(@NonNull Address[] addresses, @Nullable CallParams params)  {
        
        return inviteParticipants(nativePtr, addresses, params);
    }

    private native void oglRender(long nativePtr);
    @Override
    synchronized public void oglRender()  {
        
        oglRender(nativePtr);
    }

    private native void previewOglRender(long nativePtr);
    @Override
    synchronized public void previewOglRender()  {
        
        previewOglRender(nativePtr);
    }

    private native int removeParticipant3(long nativePtr, Call call);
    @Override
    synchronized public int removeParticipant(@NonNull Call call)  {
        
        return removeParticipant3(nativePtr, call);
    }

    private native int removeParticipant2(long nativePtr, Participant participant);
    @Override
    synchronized public int removeParticipant(@NonNull Participant participant)  {
        
        return removeParticipant2(nativePtr, participant);
    }

    private native int removeParticipant(long nativePtr, Address uri);
    @Override
    synchronized public int removeParticipant(@NonNull Address uri)  {
        
        return removeParticipant(nativePtr, uri);
    }

    private native int terminate(long nativePtr);
    @Override
    synchronized public int terminate()  {
        
        return terminate(nativePtr);
    }

    private native int updateParams(long nativePtr, ConferenceParams params);
    @Override
    synchronized public int updateParams(@NonNull ConferenceParams params)  {
        
        return updateParams(nativePtr, params);
    }

    private native void addListener(long nativePtr, ConferenceListener listener);
    @Override
    synchronized public void addListener(ConferenceListener listener)  {
        
        addListener(nativePtr, listener);
    }

    private native void removeListener(long nativePtr, ConferenceListener listener);
    @Override
    synchronized public void removeListener(ConferenceListener listener)  {
        
        removeListener(nativePtr, listener);
    }

    private native boolean unref(long ptr);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
		if (nativePtr != 0) {
			boolean destroyed = unref(nativePtr);
            if (destroyed) {
			    nativePtr = 0;
            }
		}
		super.finalize();
	}

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }

    @Override
    public String toString() {
        return "Java object [" + super.toString() + "], native pointer [" + String.format("0x%08x", nativePtr) + "]";
    }
}
